/*
ͻ
 Module       Author : David J. Yazel                                      
              Date :   December 1, 1989                                    
              Name :   I_MOUSE                                             
Ķ
 Description  These routines encompass a library of mouse interface        
              routines which use a circular buffer so that no info is lost 
              if you cannot poll fast enough, Use with I_INPUT for         
              detecting mouse presses and clicks.                          
ͼ  
*/

#include <dos.h>
#include <alloc.h>
#include <time.h>

#include "standard.h"
#include "i_mouse.h"
#include "i_lib.h"

#pragma inline

typedef void (*routine_type)();

/*  DEFINES  */

#define MOUSE_max_events  100

/* Below are the defines for this module */

#define mouse_interrupt(x,y) int86(0x33,(x),(y))
#define mouse_interrupt_extend(x,y,z) int86x(0x33,(x),(y),(z))

/* below are the variables which are used to store the queue.  There */
/* are three pieces following the classical circular buffer algorithm */
/* which works so well for this type of thing. */

static MOUSE_event_type MOUSE_event_array[MOUSE_max_events];
static unsigned MOUSE_head = 0;
static unsigned MOUSE_tail = 0;


/* Global Mouse Variables  */


char        *MOUSE_status = NULL;
boolean      MOUSE_active = FALSE;
routine_type MOUSE_old_routine = NULL;
unsigned     MOUSE_old_mask = 0;

/*  */


/* The following routine will get the mouse parameters, determine if */
/* the mouse is active and how many buttons it has. */

void MOUSE_get_parameters( active, buttons )

boolean *active;
int *buttons;

begin

     union REGS in,out;

     /* setup the call to the mouse driver for */
     /* getting the mouse parameters */

     in.x.ax = 0;
     mouse_interrupt(&in,&out);

     /* fill in the parameters */

     *active = (out.x.ax != 0);
     *buttons = (out.x.bx);

end


/* The following routine will cause the mouse cursor to appear on the */
/* screen in the current mode and defined shape.  There are no return */
/* parameters for this routine. */

void MOUSE_display_cursor()

begin

   union REGS in,out;

   /* setup the call to the mouse driver for */
   /* displaying the mouse on the screen */

   in.x.ax = 1;
   mouse_interrupt(&in,&out);


end


/* This routine will hide the mouse cursor from sight.  When this routine */
/* is invoked, the mouse cursor will disappear from the display, although */
/* the cursor can still be moved by the user */

void MOUSE_hide_cursor()

begin

   union REGS in,out;

   /* setup the call to the mouse driver to */
   /* hide the mouse cursor */

   in.x.ax = 0x02;
   mouse_interrupt(&in,&out);

end

/* the following routine will return the position of the mouse */

void MOUSE_position( xpos,ypos  )

int *xpos,*ypos;

begin


   union REGS in,out;

   /* setup the call to the mouse driver to */
   /* return the position */

   in.x.ax = 0x03;
   mouse_interrupt(&in,&out);

   /* instantiate the return variables */

   *xpos = out.x.cx;
   *ypos = out.x.dx;

end

/* The following routine will set the mouse position to the correct */
/* place on the screen.  The mouse cursor will jump to the new location */

void MOUSE_set_position( x,y )

int x,y;

begin

   union REGS in,out;

   /* setup the registers for the mouse call which will */
   /* alter the position */

   in.x.ax = 0x04;
   in.x.cx = x;
   in.x.dx = y;
   mouse_interrupt(&in,&out);

end

/* the following routine will return information about the mouse button */
/* specified.  */

void MOUSE_button_press_info( button, status,count,x,y )

int button;   /* requesting information on which button */
int *status;  /* whether the button is currently up or down */
int *count;   /* number of times the button has been pressed */
int *x,*y;    /* coordinates of the last button press */

begin

   union REGS in,out;

   /* setup the registers for the call to */
   /* the mouse interrupt which will return the requested information */

   in.x.ax = 0x05;
   in.x.bx = button;
   mouse_interrupt(&in,&out);

   /* now extract the information from the registers and place */
   /* it into the variables provided */

   *status = out.x.ax;
   *count  = out.x.bx;
   *x      = out.x.cx;
   *y      = out.x.dx;


end

/* the following routine will return information about the mouse button */
/* specified.  */

void MOUSE_button_release_info( button, status,count,x,y )

int button;   /* requesting information on which button */
int *status;  /* whether the button is currently up or down */
int *count;   /* number of times the button has been pressed */
int *x,*y;    /* coordinates of the last button press */

begin

   union REGS in,out;

   /* setup the registers for the call to */
   /* the mouse interrupt which will return the requested information */

   in.x.ax = 0x06;
   in.x.bx = button;
   mouse_interrupt(&in,&out);

   /* now extract the information from the registers and place */
   /* it into the variables provided */

   *status = out.x.ax;
   *count  = out.x.bx;
   *x      = out.x.cx;
   *y      = out.x.dx;


end


/* the following routine combines mouse functions 7 and 8 to allow the */
/* caller to specify the mouse bounds; the region which the mouse can */
/* be moved */

void MOUSE_set_bounds( tx,ty,bx,by )

int tx,ty,bx,by;

begin

   union REGS in,out;

   /* setup the registers for the mouse call which will */
   /* constrain the x bounds */

   in.x.ax = 0x07;
   in.x.cx = tx;
   in.x.dx = bx;
   mouse_interrupt(&in,&out);

   /* setup the registers for the mouse call which will */
   /* constrain the y bounds */

   in.x.ax = 0x08;
   in.x.cx = ty;
   in.x.dx = by;
   mouse_interrupt(&in,&out);

end

/* The following routine is used to define the shape of the graphics */
/* cursor.  */

void MOUSE_set_graphics_cursor( hotx,hoty,mask )

int hotx,hoty;
int mask[32];

begin


    union REGS in,out;
    struct SREGS ins;

    /* setup the call to the mouse driver */
    /* to set the new cursor shape */

    in.x.ax = 0x09;
    in.x.bx = hotx;
    in.x.cx = hoty;
    in.x.dx = FP_OFF(mask);
    ins.es  = FP_SEG(mask);
    mouse_interrupt_extend(&in,&out,&ins);

end


/* The following routine will return the number of bytes necessary to */
/* save the current mouse state.  This is usefull for allocation of */
/* dynamic memory to save the current mouse state. */

int MOUSE_get_status_size()

begin

   union REGS in,out;

   /* setup the registers for the call to the mouse driver */
   /* which will return the storage size of the setup */

   in.x.ax = 0x15;
   mouse_interrupt(&in,&out);
   return (out.x.bx);

end

/* the following routine will save the current state of the mouse */
/* driver to the buffer specified.  The amount of bytes this needs */
/* is variable and should be determined using the previously defined */
/* function. */

void MOUSE_save_status( buffer )

char *buffer;

begin

   union REGS in,out;
   struct SREGS ins;

   /* setup the registers for this mouse mouse function */

   in.x.ax = 0x16;
   in.x.dx = FP_OFF( buffer );
   ins.es  = FP_SEG( buffer );
   mouse_interrupt_extend(&in,&out,&ins);

end


/* The following routine will retore the saved mouse status.  The status */
/* should have been originally saved using the previous function. */

void MOUSE_restore_status( buffer )

char *buffer;

begin

   union REGS in,out;
   struct SREGS ins;

   /* setup the registers for this mouse mouse function */

   in.x.ax = 0x17;
   in.x.dx = FP_OFF( buffer );
   ins.es  = FP_SEG( buffer );
   mouse_interrupt_extend(&in,&out,&ins);

end

/* The following routine will allow the user to define a funtion which will*/
/* called when certain events are activated by the mouse.  This allows for */
/* some background processing to occur.  The routine specified must */
/* be a interrupt type. */

void MOUSE_set_user_function( mask, routine, old_mask, old_routine )

int mask;
routine_type routine;
unsigned *old_mask;
routine_type *old_routine;

begin

   union REGS in,out;
   struct SREGS ins;

   /* setup the registers for this mouse mouse function */

   in.x.ax = 0x14;
   in.x.cx = mask;
   in.x.dx = FP_OFF( routine );
   ins.es  = FP_SEG( routine );
   mouse_interrupt_extend(&in,&out,&ins);

   /* return the old mask and routine */

   *old_mask = out.x.cx;
   *old_routine = MK_FP(ins.es,out.x.dx);

end

/*


MOUSE_event_type event;

begin

   MOUSE_head = (MOUSE_head + 1) % MOUSE_max_events;
   MOUSE_event_array[MOUSE_head] = event;

end
*/

/* The following routine is used internally by the mouse routines to handle */
/* the direct interface with the harware mouse.  This routine looks at the */
/* currently defined areas and records if the mouse is moved into an area, */
/* moved out of an area or is clicked on that area. */

static void MOUSE_area_handler()

begin

   static int unsigned func,newx,newy;
   static MOUSE_event_type event;

   /* make sure the data segment is correct for */
   /* the procedure - The current one is for the mouse handler, not ours */
   
   asm   mov	bp,DGROUP
   asm   mov	ds,bp

   /* immediately save the horizontal and vertical counts */

   func = _AX;
   newx = _CX;
   newy = _DX;

   /* load up the event record so it can be recorded */

   event.clock = clock();
   event.x = newx;
   event.y = newy;
   event.func = func;

   /* add the event to the event queue */

   MOUSE_head = (MOUSE_head + 1) % MOUSE_max_events;
   MOUSE_event_array[MOUSE_head] = event;


end /* MOUSE_area_handler */


/* This routine is what the user calls in order to start the mouse manager */
/* to controlling the mouse and managing the areas.  This routine will */
/* install the mouse hardware handler and save the mouse status for */
/* resetting when the user closes the mouse down */

void MOUSE_initialize()

begin

   unsigned active,buttons;
   unsigned size;

   /* check to see if th mouse is installed.  If it isn't then record that */
   /* the mouse is not there and then exit. */

   MOUSE_get_parameters( &active, &buttons );
   if (!active )
     begin
       MOUSE_active = FALSE;
       return;
     end
   else MOUSE_active = TRUE;

   /* If the mouse is there then save the mouse status in a dynamically */
   /* allocated region */

   size = MOUSE_get_status_size();
   MOUSE_status = malloc( size );
   if (MOUSE_status==NULL)
     begin
       MOUSE_active = FALSE;
       return;
     end
   MOUSE_save_status( MOUSE_status );

   /* now that the status is saved to a buffer, make sure the event queue */
   /* is empty and the area stack is empty */

   MOUSE_tail = 0;
   MOUSE_head = 0;

   /* now install the mouse driver for the mouse system */

   MOUSE_set_user_function( 0xff, (routine_type)MOUSE_area_handler,
                            &MOUSE_old_mask,&MOUSE_old_routine);

   /* unhide the mouse on the screen */

   MOUSE_display_cursor();

   /* All done - return to the caller */

end

/* The following routine is user to shut down the mouse driver and */
/* restore the mouse status to what it was before. */

void MOUSE_shutdown()

begin


   unsigned mask;
   routine_type routine;

   /* if the mouse routine is not active then exit */

   if (MOUSE_active)
     begin

       /* hide the cursor, remove our mouse function and then */
       /* restore the status to the same as it was before the */
       /* mouse was initialized */

       MOUSE_hide_cursor();
       MOUSE_set_user_function( MOUSE_old_mask,(routine_type)MOUSE_old_routine,
                                &mask,&routine );
       MOUSE_restore_status( MOUSE_status );

     end

end

/* this routine will check to see if there is anything in the mouse */
/* queue which can be returned */

boolean MOUSE_event()

begin

   return (MOUSE_head!=MOUSE_tail);

end


/* this routine will return a mouse event if there is one */

void MOUSE_get_event( event )

MOUSE_event_type *event;

begin


   if (MOUSE_head!=MOUSE_tail)
     begin
       MOUSE_tail = (MOUSE_tail + 1) % MOUSE_max_events;
       *event = MOUSE_event_array[MOUSE_tail];
     end


end


void MOUSE_see_event( event )

MOUSE_event_type *event;

begin

   int new_tail;

   if (MOUSE_head!=MOUSE_tail)
     begin
       new_tail = (MOUSE_tail + 1) % MOUSE_max_events;
       *event = MOUSE_event_array[new_tail];
     end


end

void MOUSE_trash_events()

begin

   MOUSE_tail = MOUSE_head;

end
